@@ -62,16 +62,17 @@
 #include <private/qqmlnullablevalue_p.h>
 #include <private/qv4identifier_p.h>
 #include <private/qflagpointer_p.h>
-#include <private/qjson_p.h>
+#include <private/qendian_p.h>
 #ifndef V4_BOOTSTRAP
 #include <private/qqmltypenamecache_p.h>
 #include <private/qqmlpropertycache_p.h>
+#include "private/qintrusivelist_p.h"
 #endif
 
 QT_BEGIN_NAMESPACE
 
 // Bump this whenever the compiler data structures change in an incompatible way.
-#define QV4_DATA_STRUCTURE_VERSION 0x11
+#define QV4_DATA_STRUCTURE_VERSION 0x13
 
 class QIODevice;
 class QQmlPropertyCache;
@@ -96,13 +97,6 @@ class CompilationUnitMapper;
 
 namespace CompiledData {
 
-typedef QJsonPrivate::q_littleendian<qint16> LEInt16;
-typedef QJsonPrivate::q_littleendian<quint16> LEUInt16;
-typedef QJsonPrivate::q_littleendian<quint32> LEUInt32;
-typedef QJsonPrivate::q_littleendian<qint32> LEInt32;
-typedef QJsonPrivate::q_littleendian<quint64> LEUInt64;
-typedef QJsonPrivate::q_littleendian<qint64> LEInt64;
-
 struct String;
 struct Function;
 struct Lookup;
@@ -122,24 +116,22 @@ struct TableIterator
     bool operator!=(const TableIterator &rhs) const { return index != rhs.index; }
 };
 
-#if defined(Q_CC_MSVC) || defined(Q_CC_GNU)
-#pragma pack(push, 1)
-#endif
-
 struct Location
 {
     union {
-        QJsonPrivate::qle_bitfield<0, 20> line;
-        QJsonPrivate::qle_bitfield<20, 12> column;
+        quint32 _dummy;
+        quint32_le_bitfield<0, 20> line;
+        quint32_le_bitfield<20, 12> column;
     };
 
-    Location() { line.val = 0; column.val = 0; }
+    Location() : _dummy(0) { }
 
     inline bool operator<(const Location &other) const {
         return line < other.line ||
                (line == other.line && column < other.column);
     }
 };
+static_assert(sizeof(Location) == 4, "Location structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct RegExp
 {
@@ -149,12 +141,14 @@ struct RegExp
         RegExp_Multiline  = 0x04
     };
     union {
-        QJsonPrivate::qle_bitfield<0, 4> flags;
-        QJsonPrivate::qle_bitfield<4, 28> stringIndex;
+        quint32 _dummy;
+        quint32_le_bitfield<0, 4> flags;
+        quint32_le_bitfield<4, 28> stringIndex;
     };
 
-    RegExp() { flags.val = 0; stringIndex.val = 0; }
+    RegExp() : _dummy(0) { }
 };
+static_assert(sizeof(RegExp) == 4, "RegExp structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct Lookup
 {
@@ -167,40 +161,46 @@ struct Lookup
     };
 
     union {
-        QJsonPrivate::qle_bitfield<0, 4> type_and_flags;
-        QJsonPrivate::qle_bitfield<4, 28> nameIndex;
+        quint32 _dummy;
+        quint32_le_bitfield<0, 4> type_and_flags;
+        quint32_le_bitfield<4, 28> nameIndex;
     };
 
-    Lookup() { type_and_flags.val = 0; nameIndex.val = 0; }
+    Lookup() : _dummy(0) { }
 };
+static_assert(sizeof(Lookup) == 4, "Lookup structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct JSClassMember
 {
     union {
-        QJsonPrivate::qle_bitfield<0, 31> nameOffset;
-        QJsonPrivate::qle_bitfield<31, 1> isAccessor;
+        quint32 _dummy;
+        quint32_le_bitfield<0, 31> nameOffset;
+        quint32_le_bitfield<31, 1> isAccessor;
     };
 
-    JSClassMember() { nameOffset = 0; isAccessor = 0; }
+    JSClassMember() : _dummy(0) { }
 };
+static_assert(sizeof(JSClassMember) == 4, "JSClassMember structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct JSClass
 {
-    LEUInt32 nMembers;
+    quint32_le nMembers;
     // JSClassMember[nMembers]
 
     static int calculateSize(int nMembers) { return (sizeof(JSClass) + nMembers * sizeof(JSClassMember) + 7) & ~7; }
 };
+static_assert(sizeof(JSClass) == 4, "JSClass structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct String
 {
-    LEInt32 size;
+    qint32_le size;
     // uint16 strdata[]
 
     static int calculateSize(const QString &str) {
         return (sizeof(String) + str.length() * sizeof(quint16) + 7) & ~0x7;
     }
 };
+static_assert(sizeof(String) == 4, "String structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 // Function is aligned on an 8-byte boundary to make sure there are no bus errors or penalties
 // for unaligned access. The ordering of the fields is also from largest to smallest.
@@ -211,29 +211,30 @@ struct Function
         HasDirectEval       = 0x2,
         UsesArgumentsObject = 0x4,
         IsNamedExpression   = 0x8,
-        HasCatchOrWith      = 0x10
+        HasCatchOrWith      = 0x10,
+        CanUseSimpleCall    = 0x20
     };
 
     // Absolute offset into file where the code for this function is located. Only used when the function
     // is serialized.
-    LEUInt64 codeOffset;
-    LEUInt64 codeSize;
+    quint64_le codeOffset;
+    quint64_le codeSize;
 
-    LEUInt32 nameIndex;
-    LEUInt32 nFormals;
-    LEUInt32 formalsOffset;
-    LEUInt32 nLocals;
-    LEUInt32 localsOffset;
-    LEUInt32 nInnerFunctions;
+    quint32_le nameIndex;
+    quint32_le nFormals;
+    quint32_le formalsOffset;
+    quint32_le nLocals;
+    quint32_le localsOffset;
+    quint32_le nInnerFunctions;
     Location location;
 
     // Qml Extensions Begin
-    LEUInt32 nDependingIdObjects;
-    LEUInt32 dependingIdObjectsOffset; // Array of resolved ID objects
-    LEUInt32 nDependingContextProperties;
-    LEUInt32 dependingContextPropertiesOffset; // Array of int pairs (property index and notify index)
-    LEUInt32 nDependingScopeProperties;
-    LEUInt32 dependingScopePropertiesOffset; // Array of int pairs (property index and notify index)
+    quint32_le nDependingIdObjects;
+    quint32_le dependingIdObjectsOffset; // Array of resolved ID objects
+    quint32_le nDependingContextProperties;
+    quint32_le dependingContextPropertiesOffset; // Array of int pairs (property index and notify index)
+    quint32_le nDependingScopeProperties;
+    quint32_le dependingScopePropertiesOffset; // Array of int pairs (property index and notify index)
     // Qml Extensions End
 
 //    quint32 formalsIndex[nFormals]
@@ -243,16 +244,18 @@ struct Function
 
     // Keep all unaligned data at the end
     quint8 flags;
+    quint8 padding1;
+    quint16_le padding2;
 
-    const LEUInt32 *formalsTable() const { return reinterpret_cast<const LEUInt32 *>(reinterpret_cast<const char *>(this) + formalsOffset); }
-    const LEUInt32 *localsTable() const { return reinterpret_cast<const LEUInt32 *>(reinterpret_cast<const char *>(this) + localsOffset); }
-    const LEUInt32 *qmlIdObjectDependencyTable() const { return reinterpret_cast<const LEUInt32 *>(reinterpret_cast<const char *>(this) + dependingIdObjectsOffset); }
-    const LEUInt32 *qmlContextPropertiesDependencyTable() const { return reinterpret_cast<const LEUInt32 *>(reinterpret_cast<const char *>(this) + dependingContextPropertiesOffset); }
-    const LEUInt32 *qmlScopePropertiesDependencyTable() const { return reinterpret_cast<const LEUInt32 *>(reinterpret_cast<const char *>(this) + dependingScopePropertiesOffset); }
+    const quint32_le *formalsTable() const { return reinterpret_cast<const quint32_le *>(reinterpret_cast<const char *>(this) + formalsOffset); }
+    const quint32_le *localsTable() const { return reinterpret_cast<const quint32_le *>(reinterpret_cast<const char *>(this) + localsOffset); }
+    const quint32_le *qmlIdObjectDependencyTable() const { return reinterpret_cast<const quint32_le *>(reinterpret_cast<const char *>(this) + dependingIdObjectsOffset); }
+    const quint32_le *qmlContextPropertiesDependencyTable() const { return reinterpret_cast<const quint32_le *>(reinterpret_cast<const char *>(this) + dependingContextPropertiesOffset); }
+    const quint32_le *qmlScopePropertiesDependencyTable() const { return reinterpret_cast<const quint32_le *>(reinterpret_cast<const char *>(this) + dependingScopePropertiesOffset); }
 
     // --- QQmlPropertyCacheCreator interface
-    const LEUInt32 *formalsBegin() const { return formalsTable(); }
-    const LEUInt32 *formalsEnd() const { return formalsTable() + nFormals; }
+    const quint32_le *formalsBegin() const { return formalsTable(); }
+    const quint32_le *formalsEnd() const { return formalsTable() + nFormals; }
     // ---
 
     inline bool hasQmlDependencies() const { return nDependingIdObjects > 0 || nDependingContextProperties > 0 || nDependingScopeProperties > 0; }
@@ -261,17 +264,19 @@ struct Function
         return (sizeof(Function) + (nFormals + nLocals + nInnerfunctions + nIdObjectDependencies + 2 * nPropertyDependencies) * sizeof(quint32) + 7) & ~0x7;
     }
 };
+static_assert(sizeof(Function) == 72, "Function structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 // Qml data structures
 
 struct Q_QML_EXPORT TranslationData {
-    LEUInt32 commentIndex;
-    LEInt32 number;
+    quint32_le commentIndex;
+    qint32_le number;
 };
+static_assert(sizeof(TranslationData) == 8, "TranslationData structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct Q_QML_PRIVATE_EXPORT Binding
 {
-    LEUInt32 propertyNameIndex;
+    quint32_le propertyNameIndex;
 
     enum ValueType : unsigned int {
         Type_Invalid,
@@ -299,21 +304,23 @@ struct Q_QML_PRIVATE_EXPORT Binding
     };
 
     union {
-        QJsonPrivate::qle_bitfield<0, 16> flags;
-        QJsonPrivate::qle_bitfield<16, 16> type;
+        quint32_le_bitfield<0, 16> flags;
+        quint32_le_bitfield<16, 16> type;
     };
     union {
         bool b;
         quint64 doubleValue; // do not access directly, needs endian protected access
-        LEUInt32 compiledScriptIndex; // used when Type_Script
-        LEUInt32 objectIndex;
+        quint32_le compiledScriptIndex; // used when Type_Script
+        quint32_le objectIndex;
         TranslationData translationData; // used when Type_Translation
     } value;
-    LEUInt32 stringIndex; // Set for Type_String, Type_Translation and Type_Script (the latter because of script strings)
+    quint32_le stringIndex; // Set for Type_String, Type_Translation and Type_Script (the latter because of script strings)
 
     Location location;
     Location valueLocation;
 
+    quint32_le padding;
+
     bool isValueBinding() const
     {
         if (type == Type_AttachedProperty
@@ -363,7 +370,8 @@ struct Q_QML_PRIVATE_EXPORT Binding
 
     static QString escapedString(const QString &string);
 
-    bool evaluatesToString() const { return type == Type_String || type == Type_Translation || type == Type_TranslationById; }
+    bool containsTranslations() const { return type == Type_Translation || type == Type_TranslationById; }
+    bool evaluatesToString() const { return type == Type_String || containsTranslations(); }
 
     QString valueAsString(const Unit *unit) const;
     QString valueAsScriptString(const Unit *unit) const;
@@ -392,18 +400,51 @@ struct Q_QML_PRIVATE_EXPORT Binding
 
 };
 
-struct Parameter
+static_assert(sizeof(Binding) == 32, "Binding structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
+
+struct EnumValue
 {
-    LEUInt32 nameIndex;
-    LEUInt32 type;
-    LEUInt32 customTypeNameIndex;
+    quint32_le nameIndex;
+    qint32_le value;
     Location location;
 };
 
+struct Enum
+{
+    quint32_le nameIndex;
+    quint32_le nEnumValues;
+    Location location;
+
+    const EnumValue *enumValueAt(int idx) const {
+        return reinterpret_cast<const EnumValue*>(this + 1) + idx;
+    }
+
+    static int calculateSize(int nEnumValues) {
+        return (sizeof(Enum)
+                + nEnumValues * sizeof(EnumValue)
+                + 7) & ~0x7;
+    }
+
+    // --- QQmlPropertyCacheCreatorInterface
+    const EnumValue *enumValuesBegin() const { return enumValueAt(0); }
+    const EnumValue *enumValuesEnd() const { return enumValueAt(nEnumValues); }
+    int enumValueCount() const { return nEnumValues; }
+    // ---
+};
+
+struct Parameter
+{
+    quint32_le nameIndex;
+    quint32_le type;
+    quint32_le customTypeNameIndex;
+    Location location;
+};
+static_assert(sizeof(Parameter) == 16, "Parameter structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
+
 struct Signal
 {
-    LEUInt32 nameIndex;
-    LEUInt32 nParameters;
+    quint32_le nameIndex;
+    quint32_le nParameters;
     Location location;
     // Parameter parameters[1];
 
@@ -423,6 +464,7 @@ struct Signal
     int parameterCount() const { return nParameters; }
     // ---
 };
+static_assert(sizeof(Signal) == 12, "Signal structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct Property
 {
@@ -435,14 +477,15 @@ struct Property
         IsReadOnly = 0x1
     };
 
-    LEUInt32 nameIndex;
+    quint32_le nameIndex;
     union {
-        QJsonPrivate::qle_bitfield<0, 31> type;
-        QJsonPrivate::qle_bitfield<31, 1> flags; // readonly
+        quint32_le_bitfield<0, 31> type;
+        quint32_le_bitfield<31, 1> flags; // readonly
     };
-    LEUInt32 customTypeNameIndex; // If type >= Custom
+    quint32_le customTypeNameIndex; // If type >= Custom
     Location location;
 };
+static_assert(sizeof(Property) == 16, "Property structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct Alias {
     enum Flags : unsigned int {
@@ -451,18 +494,18 @@ struct Alias {
         AliasPointsToPointerObject = 0x4
     };
     union {
-        QJsonPrivate::qle_bitfield<0, 29> nameIndex;
-        QJsonPrivate::qle_bitfield<29, 3> flags;
+        quint32_le_bitfield<0, 29> nameIndex;
+        quint32_le_bitfield<29, 3> flags;
     };
     union {
-        LEUInt32 idIndex; // string index
-        QJsonPrivate::qle_bitfield<0, 31> targetObjectId; // object id index (in QQmlContextData::idValues)
-        QJsonPrivate::qle_bitfield<31, 1> aliasToLocalAlias;
+        quint32_le idIndex; // string index
+        quint32_le_bitfield<0, 31> targetObjectId; // object id index (in QQmlContextData::idValues)
+        quint32_le_bitfield<31, 1> aliasToLocalAlias;
     };
     union {
-        LEUInt32 propertyNameIndex; // string index
-        LEInt32 encodedMetaPropertyIndex;
-        LEUInt32 localAliasIndex; // index in list of aliases local to the object (if targetObjectId == objectId)
+        quint32_le propertyNameIndex; // string index
+        qint32_le encodedMetaPropertyIndex;
+        quint32_le localAliasIndex; // index in list of aliases local to the object (if targetObjectId == objectId)
     };
     Location location;
     Location referenceLocation;
@@ -472,6 +515,7 @@ struct Alias {
         return encodedMetaPropertyIndex == -1;
     }
 };
+static_assert(sizeof(Alias) == 20, "Alias structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct Object
 {
@@ -485,26 +529,28 @@ struct Object
     // Depending on the use, this may be the type name to instantiate before instantiating this
     // object. For grouped properties the type name will be empty and for attached properties
     // it will be the name of the attached type.
-    LEUInt32 inheritedTypeNameIndex;
-    LEUInt32 idNameIndex;
+    quint32_le inheritedTypeNameIndex;
+    quint32_le idNameIndex;
     union {
-        QJsonPrivate::qle_bitfield<0, 15> flags;
-        QJsonPrivate::qle_bitfield<15, 1> defaultPropertyIsAlias;
-        QJsonPrivate::qle_signedbitfield<16, 16> id;
+        quint32_le_bitfield<0, 15> flags;
+        quint32_le_bitfield<15, 1> defaultPropertyIsAlias;
+        qint32_le_bitfield<16, 16> id;
     };
-    LEInt32 indexOfDefaultPropertyOrAlias; // -1 means no default property declared in this object
-    LEUInt32 nFunctions;
-    LEUInt32 offsetToFunctions;
-    LEUInt32 nProperties;
-    LEUInt32 offsetToProperties;
-    LEUInt32 nAliases;
-    LEUInt32 offsetToAliases;
-    LEUInt32 nSignals;
-    LEUInt32 offsetToSignals; // which in turn will be a table with offsets to variable-sized Signal objects
-    LEUInt32 nBindings;
-    LEUInt32 offsetToBindings;
-    LEUInt32 nNamedObjectsInComponent;
-    LEUInt32 offsetToNamedObjectsInComponent;
+    qint32_le indexOfDefaultPropertyOrAlias; // -1 means no default property declared in this object
+    quint32_le nFunctions;
+    quint32_le offsetToFunctions;
+    quint32_le nProperties;
+    quint32_le offsetToProperties;
+    quint32_le nAliases;
+    quint32_le offsetToAliases;
+    quint32_le nEnums;
+    quint32_le offsetToEnums; // which in turn will be a table with offsets to variable-sized Enum objects
+    quint32_le nSignals;
+    quint32_le offsetToSignals; // which in turn will be a table with offsets to variable-sized Signal objects
+    quint32_le nBindings;
+    quint32_le offsetToBindings;
+    quint32_le nNamedObjectsInComponent;
+    quint32_le offsetToNamedObjectsInComponent;
     Location location;
     Location locationOfIdProperty;
 //    Function[]
@@ -512,12 +558,13 @@ struct Object
 //    Signal[]
 //    Binding[]
 
-    static int calculateSizeExcludingSignals(int nFunctions, int nProperties, int nAliases, int nSignals, int nBindings, int nNamedObjectsInComponent)
+    static int calculateSizeExcludingSignalsAndEnums(int nFunctions, int nProperties, int nAliases, int nEnums, int nSignals, int nBindings, int nNamedObjectsInComponent)
     {
         return ( sizeof(Object)
                  + nFunctions * sizeof(quint32)
                  + nProperties * sizeof(Property)
                  + nAliases * sizeof(Alias)
+                 + nEnums * sizeof(quint32)
                  + nSignals * sizeof(quint32)
                  + nBindings * sizeof(Binding)
                  + nNamedObjectsInComponent * sizeof(int)
@@ -525,9 +572,9 @@ struct Object
                ) & ~0x7;
     }
 
-    const LEUInt32 *functionOffsetTable() const
+    const quint32_le *functionOffsetTable() const
     {
-        return reinterpret_cast<const LEUInt32*>(reinterpret_cast<const char *>(this) + offsetToFunctions);
+        return reinterpret_cast<const quint32_le*>(reinterpret_cast<const char *>(this) + offsetToFunctions);
     }
 
     const Property *propertyTable() const
@@ -545,21 +592,29 @@ struct Object
         return reinterpret_cast<const Binding*>(reinterpret_cast<const char *>(this) + offsetToBindings);
     }
 
+    const Enum *enumAt(int idx) const
+    {
+        const quint32_le *offsetTable = reinterpret_cast<const quint32_le*>((reinterpret_cast<const char *>(this)) + offsetToEnums);
+        const quint32_le offset = offsetTable[idx];
+        return reinterpret_cast<const Enum*>(reinterpret_cast<const char*>(this) + offset);
+    }
+
     const Signal *signalAt(int idx) const
     {
-        const LEUInt32 *offsetTable = reinterpret_cast<const LEUInt32*>((reinterpret_cast<const char *>(this)) + offsetToSignals);
-        const LEUInt32 offset = offsetTable[idx];
+        const quint32_le *offsetTable = reinterpret_cast<const quint32_le*>((reinterpret_cast<const char *>(this)) + offsetToSignals);
+        const quint32_le offset = offsetTable[idx];
         return reinterpret_cast<const Signal*>(reinterpret_cast<const char*>(this) + offset);
     }
 
-    const LEUInt32 *namedObjectsInComponentTable() const
+    const quint32_le *namedObjectsInComponentTable() const
     {
-        return reinterpret_cast<const LEUInt32*>(reinterpret_cast<const char *>(this) + offsetToNamedObjectsInComponent);
+        return reinterpret_cast<const quint32_le*>(reinterpret_cast<const char *>(this) + offsetToNamedObjectsInComponent);
     }
 
     // --- QQmlPropertyCacheCreator interface
     int propertyCount() const { return nProperties; }
     int aliasCount() const { return nAliases; }
+    int enumCount() const { return nEnums; }
     int signalCount() const { return nSignals; }
     int functionCount() const { return nFunctions; }
 
@@ -572,6 +627,10 @@ struct Object
     const Alias *aliasesBegin() const { return aliasTable(); }
     const Alias *aliasesEnd() const { return aliasTable() + nAliases; }
 
+    typedef TableIterator<Enum, Object, &Object::enumAt> EnumIterator;
+    EnumIterator enumsBegin() const { return EnumIterator(this, 0); }
+    EnumIterator enumsEnd() const { return EnumIterator(this, nEnums); }
+
     typedef TableIterator<Signal, Object, &Object::signalAt> SignalIterator;
     SignalIterator signalsBegin() const { return SignalIterator(this, 0); }
     SignalIterator signalsEnd() const { return SignalIterator(this, nSignals); }
@@ -579,6 +638,7 @@ struct Object
     int namedObjectsInComponentCount() const { return nNamedObjectsInComponent; }
     // ---
 };
+static_assert(sizeof(Object) == 80, "Object structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct Import
 {
@@ -587,18 +647,19 @@ struct Import
         ImportFile = 0x2,
         ImportScript = 0x3
     };
-    LEUInt32 type;
+    quint32_le type;
 
-    LEUInt32 uriIndex;
-    LEUInt32 qualifierIndex;
+    quint32_le uriIndex;
+    quint32_le qualifierIndex;
 
-    LEInt32 majorVersion;
-    LEInt32 minorVersion;
+    qint32_le majorVersion;
+    qint32_le minorVersion;
 
     Location location;
 
     Import() { type = 0; uriIndex = 0; qualifierIndex = 0; majorVersion = 0; minorVersion = 0; }
 };
+static_assert(sizeof(Import) == 24, "Import structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 static const char magic_str[] = "qv4cdata";
 
@@ -606,17 +667,17 @@ struct Unit
 {
     // DO NOT CHANGE THESE FIELDS EVER
     char magic[8];
-    LEUInt32 version;
-    LEUInt32 qtVersion;
-    LEInt64 sourceTimeStamp;
-    LEUInt32 unitSize; // Size of the Unit and any depending data.
+    quint32_le version;
+    quint32_le qtVersion;
+    qint64_le sourceTimeStamp;
+    quint32_le unitSize; // Size of the Unit and any depending data.
     // END DO NOT CHANGE THESE FIELDS EVER
 
     char md5Checksum[16]; // checksum of all bytes following this field.
     void generateChecksum();
 
-    LEUInt32 architectureIndex; // string index to QSysInfo::buildAbi()
-    LEUInt32 codeGeneratorIndex;
+    quint32_le architectureIndex; // string index to QSysInfo::buildAbi()
+    quint32_le codeGeneratorIndex;
     char dependencyMD5Checksum[16];
 
     enum : unsigned int {
@@ -628,36 +689,35 @@ struct Unit
         ContainsMachineCode = 0x20, // used to determine if we need to mmap with execute permissions
         PendingTypeCompilation = 0x40 // the QML data structures present are incomplete and require type compilation
     };
-    LEUInt32 flags;
-    LEUInt32 stringTableSize;
-    LEUInt32 offsetToStringTable;
-    LEUInt32 functionTableSize;
-    LEUInt32 offsetToFunctionTable;
-    LEUInt32 lookupTableSize;
-    LEUInt32 offsetToLookupTable;
-    LEUInt32 regexpTableSize;
-    LEUInt32 offsetToRegexpTable;
-    LEUInt32 constantTableSize;
-    LEUInt32 offsetToConstantTable;
-    LEUInt32 jsClassTableSize;
-    LEUInt32 offsetToJSClassTable;
-    LEInt32 indexOfRootFunction;
-    LEUInt32 sourceFileIndex;
+    quint32_le flags;
+    quint32_le stringTableSize;
+    quint32_le offsetToStringTable;
+    quint32_le functionTableSize;
+    quint32_le offsetToFunctionTable;
+    quint32_le lookupTableSize;
+    quint32_le offsetToLookupTable;
+    quint32_le regexpTableSize;
+    quint32_le offsetToRegexpTable;
+    quint32_le constantTableSize;
+    quint32_le offsetToConstantTable;
+    quint32_le jsClassTableSize;
+    quint32_le offsetToJSClassTable;
+    qint32_le indexOfRootFunction;
+    quint32_le sourceFileIndex;
 
     /* QML specific fields */
-    LEUInt32 nImports;
-    LEUInt32 offsetToImports;
-    LEUInt32 nObjects;
-    LEUInt32 offsetToObjects;
-    LEUInt32 indexOfRootObject;
+    quint32_le nImports;
+    quint32_le offsetToImports;
+    quint32_le nObjects;
+    quint32_le offsetToObjects;
 
     const Import *importAt(int idx) const {
         return reinterpret_cast<const Import*>((reinterpret_cast<const char *>(this)) + offsetToImports + idx * sizeof(Import));
     }
 
     const Object *objectAt(int idx) const {
-        const LEUInt32 *offsetTable = reinterpret_cast<const LEUInt32*>((reinterpret_cast<const char *>(this)) + offsetToObjects);
-        const LEUInt32 offset = offsetTable[idx];
+        const quint32_le *offsetTable = reinterpret_cast<const quint32_le*>((reinterpret_cast<const char *>(this)) + offsetToObjects);
+        const quint32_le offset = offsetTable[idx];
         return reinterpret_cast<const Object*>(reinterpret_cast<const char*>(this) + offset);
     }
 
@@ -667,8 +727,8 @@ struct Unit
     /* end QML specific fields*/
 
     QString stringAt(int idx) const {
-        const LEUInt32 *offsetTable = reinterpret_cast<const LEUInt32*>((reinterpret_cast<const char *>(this)) + offsetToStringTable);
-        const LEUInt32 offset = offsetTable[idx];
+        const quint32_le *offsetTable = reinterpret_cast<const quint32_le*>((reinterpret_cast<const char *>(this)) + offsetToStringTable);
+        const quint32_le offset = offsetTable[idx];
         const String *str = reinterpret_cast<const String*>(reinterpret_cast<const char *>(this) + offset);
         if (str->size == 0)
             return QString();
@@ -680,7 +740,7 @@ struct Unit
         //            return QString::fromRawData(characters, str->size);
         return QString(characters, str->size);
 #else
-        const LEUInt16 *characters = reinterpret_cast<const LEUInt16 *>(str + 1);
+        const quint16_le *characters = reinterpret_cast<const quint16_le *>(str + 1);
         QString qstr(str->size, Qt::Uninitialized);
         QChar *ch = qstr.data();
         for (int i = 0; i < str->size; ++i)
@@ -689,11 +749,11 @@ struct Unit
 #endif
     }
 
-    const LEUInt32 *functionOffsetTable() const { return reinterpret_cast<const LEUInt32*>((reinterpret_cast<const char *>(this)) + offsetToFunctionTable); }
+    const quint32_le *functionOffsetTable() const { return reinterpret_cast<const quint32_le*>((reinterpret_cast<const char *>(this)) + offsetToFunctionTable); }
 
     const Function *functionAt(int idx) const {
-        const LEUInt32 *offsetTable = functionOffsetTable();
-        const LEUInt32 offset = offsetTable[idx];
+        const quint32_le *offsetTable = functionOffsetTable();
+        const quint32_le offset = offsetTable[idx];
         return reinterpret_cast<const Function*>(reinterpret_cast<const char *>(this) + offset);
     }
 
@@ -701,13 +761,13 @@ struct Unit
     const RegExp *regexpAt(int index) const {
         return reinterpret_cast<const RegExp*>(reinterpret_cast<const char *>(this) + offsetToRegexpTable + index * sizeof(RegExp));
     }
-    const LEUInt64 *constants() const {
-        return reinterpret_cast<const LEUInt64*>(reinterpret_cast<const char *>(this) + offsetToConstantTable);
+    const quint64_le *constants() const {
+        return reinterpret_cast<const quint64_le*>(reinterpret_cast<const char *>(this) + offsetToConstantTable);
     }
 
     const JSClassMember *jsClassAt(int idx, int *nMembers) const {
-        const LEUInt32 *offsetTable = reinterpret_cast<const LEUInt32 *>(reinterpret_cast<const char *>(this) + offsetToJSClassTable);
-        const LEUInt32 offset = offsetTable[idx];
+        const quint32_le *offsetTable = reinterpret_cast<const quint32_le *>(reinterpret_cast<const char *>(this) + offsetToJSClassTable);
+        const quint32_le offset = offsetTable[idx];
         const char *ptr = reinterpret_cast<const char *>(this) + offset;
         const JSClass *klass = reinterpret_cast<const JSClass *>(ptr);
         *nMembers = klass->nMembers;
@@ -715,9 +775,7 @@ struct Unit
     }
 };
 
-#if defined(Q_CC_MSVC) || defined(Q_CC_GNU)
-#pragma pack(pop)
-#endif
+static_assert(sizeof(Unit) == 144, "Unit structure needs to have the expected size to be binary compatible on disk when generated by host compiler and loaded by target");
 
 struct TypeReference
 {
@@ -823,7 +881,9 @@ struct Q_QML_PRIVATE_EXPORT CompilationUnit : public CompilationUnitBase, public
     virtual QV4::CompiledData::Unit *createUnitData(QmlIR::Document *irDocument);
 
 #ifndef V4_BOOTSTRAP
+    QIntrusiveListNode nextCompilationUnit;
     ExecutionEngine *engine;
+    QQmlEnginePrivate *qmlEngine; // only used in QML environment for composite types, not in plain QJSEngine case.
 
     QString fileName() const { return data->stringAt(data->sourceFileIndex); }
     QUrl url() const { if (m_url.isNull) m_url = QUrl(fileName()); return m_url; }
@@ -836,7 +896,7 @@ struct Q_QML_PRIVATE_EXPORT CompilationUnit : public CompilationUnitBase, public
 
     // QML specific fields
     QQmlPropertyCacheVector propertyCaches;
-    QQmlPropertyCache *rootPropertyCache() const { return propertyCaches.at(data->indexOfRootObject); }
+    QQmlPropertyCache *rootPropertyCache() const { return propertyCaches.at(/*root object*/0); }
 
     QQmlRefPointer<QQmlTypeNameCache> typeNameCache;
 
@@ -853,7 +913,7 @@ struct Q_QML_PRIVATE_EXPORT CompilationUnit : public CompilationUnitBase, public
     // pointers either to data->constants() or little-endian memory copy.
     const Value* constants;
 
-    void finalize(QQmlEnginePrivate *engine);
+    void finalizeCompositeType(QQmlEnginePrivate *qmlEngine);
 
     int totalBindingsCount; // Number of bindings used in this type
     int totalParserStatusCount; // Number of instantiated types that are QQmlParserStatus subclasses
@@ -873,7 +933,6 @@ struct Q_QML_PRIVATE_EXPORT CompilationUnit : public CompilationUnitBase, public
     // --- interface for QQmlPropertyCacheCreator
     typedef Object CompiledObject;
     int objectCount() const { return data->nObjects; }
-    int rootObjectIndex() const { return data->indexOfRootObject; }
     const Object *objectAt(int index) const { return data->objectAt(index); }
     QString stringAt(int index) const { return data->stringAt(index); }
 
@@ -896,7 +955,7 @@ struct Q_QML_PRIVATE_EXPORT CompilationUnit : public CompilationUnitBase, public
     QV4::Function *linkToEngine(QV4::ExecutionEngine *engine);
     void unlink();
 
-    void markObjects(QV4::ExecutionEngine *e);
+    void markObjects(MarkStack *markStack);
 
     void destroy() Q_DECL_OVERRIDE;
 
@@ -923,13 +982,12 @@ protected:
 struct ResolvedTypeReference
 {
     ResolvedTypeReference()
-        : type(0)
-        , majorVersion(0)
+        : majorVersion(0)
         , minorVersion(0)
         , isFullyDynamicType(false)
     {}
 
-    QQmlType *type;
+    QQmlType type;
     QQmlRefPointer<QQmlPropertyCache> typePropertyCache;
     QQmlRefPointer<QV4::CompiledData::CompilationUnit> compilationUnit;
 
